/*-----------------------------------------------------------------------------
 * Name:    Touch_ADS7843.c
 * Purpose: Touchscreen ADS7843 interface
 * Rev.:    1.00
 *----------------------------------------------------------------------------*/

/* Copyright (c) 2013 - 2014 ARM LIMITED

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
   ---------------------------------------------------------------------------*/

#include <stdbool.h>
#include "cmsis_os.h"                   // ARM::CMSIS:RTOS:Keil RTX
#include "Board_Touch.h"                // ::Board Support:Touchscreen
#include "sam4e.h"                      // Device header

/* Touch screen IRQ & Busy pin definition */
#define ADS7843_PIN_IRQ     (PIO_PA16)
#define ADS7843_PIN_BUSY    (PIO_PA17)

/* Touch screen control register bit definitions */
#define ADS7843_CTRL_PD     (3 << 0)                /* no power down */
#define ADS7843_CTRL_Y      (1 << 4)                /* y position    */
#define ADS7843_CTRL_X      (5 << 4)                /* x position    */
#define ADS7843_CTRL_START  (1 << 7)                /* start bit     */

/* Touch screen commands */
#define CMD_Y_POSITION    (ADS7843_CTRL_START | ADS7843_CTRL_Y | ADS7843_CTRL_PD)
#define CMD_X_POSITION    (ADS7843_CTRL_START | ADS7843_CTRL_X | ADS7843_CTRL_PD)
#define CMD_ENABLE_PENIRQ (ADS7843_CTRL_START | ADS7843_CTRL_X)


/* SPI pins
      CS         PA11         GPIO
      MISO       PA12         PIO_PA12A_MISO
      MOSI       PA13         PIO_PA13A_MOSI
      CLK        PA14         PIO_PA14A_SPCK
 */
#define SPI_SS      (PIO_PA11)
#define SPI_MISO    (PIO_PA12A_MISO)
#define SPI_MOSI    (PIO_PA13A_MOSI)
#define SPI_CLK     (PIO_PA14A_SPCK)

#define SPI_CS      0
#define SPI_PCS    (~(1<<SPI_CS) & 0x0F)         /* set PCS according CS */


/**
  \fn          uint8_t SPI_Send (uint8_t outb)
  \brief       Write and Read a byte on SPI interface
  \param[in]   outb    Byte to write
  \returns     Byte read
*/
static uint8_t SPI_Send (uint8_t outb) {

  /* Wait until TDRE is set. */
  SPI->SPI_TDR = outb;

  /* Wait until RDRF is set. */
  while (!(SPI->SPI_SR & SPI_SR_RDRF));
  return ((uint8_t)SPI->SPI_RDR);
}


/**
  \fn          bool Touch_WriteCmd (uint8_t cmd, uint32_t *result)
  \brief       Write command to Touchscreen controller
  \param[in]   cmd    Command to write
  \param[in]   result Pointer to result data
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
static bool Touch_WriteCmd (uint8_t cmd, uint32_t *result) {
  volatile uint8_t bufferRX[3];
  uint32_t uResult = 0;

  PIOA->PIO_CODR = SPI_SS;                 /* enable chip select */

  bufferRX[0] = SPI_Send(cmd);
  bufferRX[1] = SPI_Send(0);
  bufferRX[2] = SPI_Send(0);

  PIOA->PIO_SODR = SPI_SS;                 /* disable chip select */

  uResult = ((uint32_t)bufferRX[1] << 8) | (uint32_t)bufferRX[2] ;
  *result = uResult >> 4;

  return 0;
}


/**
  \fn          bool Touch_Pressed (void)
  \brief       get touch state
  \returns
   - \b  0: Not pressed
   - \b  1: pressed
*/
static bool Touch_Pressed (void) {

  return((PIOA->PIO_PDSR & ADS7843_PIN_IRQ) == 0);
}


/**
  \fn          int32_t Touch_Initialize (void)
  \brief       Initialize touchscreen
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Touch_Initialize (void) {
  uint32_t tick;
  uint32_t touchResult;

  SystemCoreClockUpdate();

  PMC->PMC_WPMR = PMC_WPMR_WPKEY_PASSWD;    /* Disable PMC write protect  */
  PMC->PMC_PCER0 = ((1UL << ID_PIOA) |      /* enable clock for SPI pins  */
                    (1UL << ID_SPI)   );    /* enable SPI clock           */

  PIOA->PIO_WPMR = PIO_WPMR_WPKEY_PASSWD;   /* disable PIOA write protect */
  PIOA->PIO_PUER =
  PIOA->PIO_IDR  =
  PIOA->PIO_ODR  =
  PIOA->PIO_PER  = (ADS7843_PIN_IRQ | ADS7843_PIN_BUSY);

  /* SPI_MISO, SPI_MOSI, SPI_SPCK are SPI pins */
  PIOA->PIO_ABCDSR[0] &= ~(SPI_MISO |
                           SPI_MOSI |
                           SPI_CLK   );
  PIOA->PIO_PDR        =  (SPI_MISO |
                           SPI_MOSI |
                           SPI_CLK   );

  /* SPIO_NPCS0 is GPIO, output set to high. */
  PIOA->PIO_PER  =
  PIOA->PIO_OER  =
  PIOA->PIO_SODR = SPI_SS;

  /* Initialize SPI Controller */
  SPI->SPI_CR          = SPI_CR_SWRST;
  SPI->SPI_MR          = SPI_MR_MSTR    |
                         SPI_MR_MODFDIS |
                         SPI_MR_PCS(SPI_PCS); /* Use NPCSx as chip select */

  SPI->SPI_CSR[SPI_CS] = SPI_CSR_NCPHA |
                         SPI_CSR_SCBR(SystemCoreClock/500000ul); /* 2Mhz @ 96 MHz */

  SPI->SPI_CR          = SPI_CR_SPIEN;

  tick = osKernelSysTick();                   /* Wait 10ms                     */
  while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(10000));

  Touch_WriteCmd(CMD_ENABLE_PENIRQ, &touchResult);

  return 0;
}

/**
  \fn          int32_t Touch_Uninitialize (void)
  \brief       De-initialize touchscreen
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Touch_Uninitialize (void) {

  PIOA->PIO_PER         = SPI_MISO |
                          SPI_MOSI |
                          SPI_CLK ;

  PIOA->PIO_PER         =
  PIOA->PIO_ODR         =
  PIOA->PIO_CODR        = SPI_SS;

  SPI->SPI_CR          = 0;
  SPI->SPI_MR          = 0;
  SPI->SPI_CSR[SPI_CS] = 0;

  PMC->PMC_PCDR0 = (1UL << ID_SPI);       /* disable SPI clock                */

  return 0;
}

/**
  \fn          int32_t Touch_GetState (TOUCH_STATE *pState)
  \brief       Get touchscreen state
  \param[out]  pState  pointer to TOUCH_STATE structure
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Touch_GetState (TOUCH_STATE *pState) {
  int32_t res;
  uint32_t x, y;

  pState->pressed = Touch_Pressed();

  if (pState->pressed) {
    res = Touch_WriteCmd(CMD_X_POSITION, &x);
    if (res !=0) {
      return (-1);
    }
    res = Touch_WriteCmd(CMD_Y_POSITION, &y);
    if (res !=0) {
      return (-1);
    }

    pState->x = x;
    pState->y = y;

    Touch_WriteCmd(CMD_ENABLE_PENIRQ, &x);
  }

  return 0;
}
